@@ -1,8 +1,7 @@
 #!/usr/bin/env python3
 #
-# Copyright (C) 2008 Henri Hakkinen
-#
-# Copyright (C) 2015-2016 Arun Prakash Jana <engineerarun@gmail.com>
+# Copyright © 2008 Henri Hakkinen
+# Copyright © 2015-2017 Arun Prakash Jana <engineerarun@gmail.com>
 #
 # This program is free software: you can redistribute it and/or modify
 # it under the terms of the GNU General Public License as published by
@@ -56,6 +55,7 @@ except Exception:
 logging.basicConfig(format='[%(levelname)s] %(message)s')
 logger = logging.getLogger()
 
+
 def sigint_handler(signum, frame):
     print('\nInterrupted.', file=sys.stderr)
     sys.exit(1)
@@ -65,7 +65,7 @@ signal.signal(signal.SIGINT, sigint_handler)
 
 # Constants
 
-_VERSION_ = '2.9'
+_VERSION_ = '3.1'
 
 COLORMAP = {k: '\x1b[%sm' % v for k, v in {
     'a': '30', 'b': '31', 'c': '32', 'd': '33',
@@ -80,7 +80,7 @@ COLORMAP = {k: '\x1b[%sm' % v for k, v in {
 }.items()}
 
 # Disguise as Firefox on Ubuntu
-USER_AGENT = ('Mozilla/5.0 (X11; Ubuntu; Linux x86_64; rv:48.0) Gecko/20100101 Firefox/48.0')
+USER_AGENT = ('Mozilla/5.0 (X11; Ubuntu; Linux x86_64; rv:53.0) Gecko/20100101 Firefox/53.0')
 ua = True  # User Agent is enabled by default
 
 # Self-upgrade parameters
@@ -92,7 +92,7 @@ ua = True  # User Agent is enabled by default
 #
 # before running `make install'.
 
-ENABLE_SELF_UPGRADE_MECHANISM=True
+ENABLE_SELF_UPGRADE_MECHANISM = True
 API_REPO_BASE = 'https://api.github.com/repos/jarun/googler'
 RAW_DOWNLOAD_REPO_BASE = 'https://raw.githubusercontent.com/jarun/googler'
 
@@ -121,7 +121,7 @@ def open_url(url):
         os.dup2(fd, 2)
         os.dup2(fd, 1)
     try:
-        webbrowser.open(url)
+        webbrowser.open(url, new=2)
     finally:
         if open_url.suppress_browser_output:
             os.close(fd)
@@ -148,7 +148,7 @@ def unwrap(text):
             result += '\n\n'
         elif lines[i + 1]:
             # Next line is not paragraph break, add space
-            result +=  ' '
+            result += ' '
     # Handle last line
     result += lines[-1] if lines[-1] else '\n'
     return result
@@ -157,7 +157,7 @@ def unwrap(text):
 def check_stdout_encoding():
     """Make sure stdout encoding is utf-8.
 
-    If not, print error message and intrustructions, then exit with
+    If not, print error message and instructions, then exit with
     status 1.
 
     This function is a no-op on win32 because encoding on win32 is
@@ -206,13 +206,17 @@ class TLS1_2Connection(HTTPSConnection):
 
     def connect(self, notweak=False):
         sock = socket.create_connection((self.host, self.port),
-                self.timeout, self.source_address)
+                                        self.timeout, self.source_address)
 
         # Optimizations not available on OS X
         if not notweak and sys.platform.startswith('linux'):
-            sock.setsockopt(socket.SOL_TCP, socket.TCP_DEFER_ACCEPT, 1)
-            sock.setsockopt(socket.IPPROTO_TCP, socket.TCP_QUICKACK, 1)
-            sock.setsockopt(socket.SOL_SOCKET,socket.SO_RCVBUF, 524288)
+            try:
+                sock.setsockopt(socket.SOL_TCP, socket.TCP_DEFER_ACCEPT, 1)
+                sock.setsockopt(socket.IPPROTO_TCP, socket.TCP_QUICKACK, 1)
+                sock.setsockopt(socket.SOL_SOCKET, socket.SO_RCVBUF, 524288)
+            except OSError:
+                # Doesn't work on Windows' Linux subsystem (#179)
+                logger.debug('setsockopt failed')
 
         if getattr(self, '_tunnel_host', None):
             self.sock = sock
@@ -234,11 +238,12 @@ class TLS1_2Connection(HTTPSConnection):
         # Fallback
         HTTPSConnection.connect(self)
 
+
 class GoogleUrl(object):
     """
     This class constructs the Google Search/News URL.
 
-    This class is modeled on urllib.parse.ParseResult for familiarity,
+    This class is modelled on urllib.parse.ParseResult for familiarity,
     which means it supports reading of all six attributes -- scheme,
     netloc, path, params, query, fragment -- of
     urllib.parse.ParseResult, as well as the geturl() method.
@@ -291,7 +296,7 @@ class GoogleUrl(object):
         self._num = 10
         self._start = 0
         self._keywords = []
-        self._site = None
+        self._sites = None
         self._query_dict = {
             'ie': 'UTF-8',
             'oe': 'UTF-8',
@@ -417,8 +422,8 @@ class GoogleUrl(object):
                 qd.pop('tbm', None)
         if 'num' in opts:
             self._num = opts['num']
-        if 'site' in opts:
-            self._site = opts['site']
+        if 'sites' in opts:
+            self._sites = opts['sites']
         if 'start' in opts:
             self._start = opts['start']
         if 'tld' in opts:
@@ -574,21 +579,23 @@ class GoogleUrl(object):
         qd = {}
         qd.update(self._query_dict)
         qd['num'] = self._num
-        qd['start'] = self._start
+        if self._start:
+            qd['start'] = self._start
 
         # Construct the q query
         q = ''
         keywords = self._keywords
+        sites = self._sites
         if keywords:
             if isinstance(keywords, list):
-                q += '+'.join([urllib.parse.quote_plus(kw) for kw in keywords])
+                q += '+'.join(urllib.parse.quote_plus(kw) for kw in keywords)
             else:
                 q += urllib.parse.quote_plus(keywords)
-        if self._site:
-            q += '+site:' + urllib.parse.quote_plus(self._site)
+        if sites:
+            q += '+OR'.join('+site:' + urllib.parse.quote_plus(site) for site in sites)
         qd['q'] = q
 
-        return '&'.join(['%s=%s' % (k, qd[k]) for k in sorted(qd.keys())])
+        return '&'.join('%s=%s' % (k, qd[k]) for k in sorted(qd.keys()))
 
 
 class GoogleConnectionError(Exception):
@@ -707,7 +714,7 @@ class GoogleConnection(object):
         GoogleConnectionError
             When not getting HTTP 200 even after the allowed one
             reconnection and/or one redirection, or when Google is
-            blocking query due to unsual activity.
+            blocking query due to unusual activity.
 
         Returns
         -------
@@ -732,7 +739,7 @@ class GoogleConnection(object):
         while resp.status != 200 and redirect_counter < 3:
             if resp.status in {301, 302, 303, 307, 308}:
                 redirection_url = resp.getheader('location', '')
-                if 'sorry/IndexRedirect?' in redirection_url:
+                if 'sorry/IndexRedirect?' in redirection_url or 'sorry/index?' in redirection_url:
                     raise GoogleConnectionError('Connection blocked due to unusual activity.')
                 self._redirect(redirection_url)
                 resp = self._resp
@@ -871,6 +878,7 @@ def annotate_tag(annotated_starttag_handler):
 
     return handler
 
+
 def retrieve_tag_annotation(annotated_endtag_handler):
     # See parser logic within the GoogleParser class for documentation.
     #
@@ -887,6 +895,7 @@ def retrieve_tag_annotation(annotated_endtag_handler):
 
     return handler
 
+
 class GoogleParser(html.parser.HTMLParser):
     """The members of this class parse the result
     HTML page fetched from Google server for a query.
@@ -905,11 +914,11 @@ class GoogleParser(html.parser.HTMLParser):
     #
     #   1. Tag handlers are contextual;
     #
-    #   2. Contexual starttag and endtag handlers should come in pairs
+    #   2. Contextual starttag and endtag handlers should come in pairs
     #      and have a clear hierarchy;
     #
-    #   3. Starttag handlers should only yield control to a pair of
-    #      child handlers (that is, one level down the hierachy), and
+    #   3. starttag handlers should only yield control to a pair of
+    #      child handlers (that is, one level down the hierarchy), and
     #      correspondingly, endtag handlers should only return control
     #      to the parent (that is, the pair of handlers that gave it
     #      control in the first place).
@@ -1171,7 +1180,7 @@ class GoogleParser(html.parser.HTMLParser):
 
         self.set_handlers_to('root')
 
-    ### Ignore list ###
+    # Ignore list
     IGNORE_LIST = [
         # "People also ask"
         # Sample query: VPN
@@ -1179,21 +1188,30 @@ class GoogleParser(html.parser.HTMLParser):
         {
             'tag': 'div',
             'class': 'related-question-pair'
+        },
+        # We omit Google's "smart card" results (term coined by me) by
+        # guarding against the 'g-blk' class (sample response: https://git.io/voJgB)
+        {
+            'tag': 'div',
+            'class': 'g-blk'
+        },
+        # We also guard against "smart-card" results with `--noua` option
+        {
+            'tag': 'div',
+            'class': 'hp-xpdbox'
         }
     ]
 
-    ### Tag handlers ###
+    # Tag handlers
 
     @annotate_tag
     def root_start(self, tag, attrs):
-        # We omit Google's "smart card" results (term coined by me) by
-        # guarding against the 'g-blk' class (sample response: https://git.io/voJgB)
-        if tag == 'div' and 'g' in self.classes(attrs) and 'g-blk' not in self.classes(attrs):
+        if tag == 'div' and 'g' in self.classes(attrs):
             # Initialize result field registers
             self.title = ''
             self.url = ''
             self.abstract = ''
-            self.metadata = '' # Only used for Google News
+            self.metadata = ''  # Only used for Google News
             self.sitelinks = []
 
             # Guard against sitelinks, which also have titles and
@@ -1201,7 +1219,7 @@ class GoogleParser(html.parser.HTMLParser):
             # sections" (secondary results to the same event).
             self.title_registered = False
             self.abstract_registered = False
-            self.metadata_registered = False # Only used for Google News
+            self.metadata_registered = False  # Only used for Google News
 
             self.set_handlers_to('result')
             return 'result'
@@ -1255,7 +1273,7 @@ class GoogleParser(html.parser.HTMLParser):
                     self.start_populating_textbuf(lambda text: ', ' if text == '-' else text)
                 else:
                     self.start_populating_textbuf(lambda text:
-                                              text.replace(' -', ',', 1) if ' - ' in text else text)
+                                                  text.replace(' -', ',', 1) if ' - ' in text else text)
                 return 'news_metadata'
 
             if not self.abstract_registered and tag == 'div' and 'st' in self.classes(attrs):
@@ -1283,7 +1301,7 @@ class GoogleParser(html.parser.HTMLParser):
         elif annotation == 'abstract_gservices':
             self.stop_populating_textbuf()
             self.abstract = self.pop_textbuf().replace('  ', ' ')
-            self.abstract_registered = False
+            self.abstract_registered = True
 
     @annotate_tag
     def title_start(self, tag, attrs):
@@ -1319,16 +1337,24 @@ class GoogleParser(html.parser.HTMLParser):
 
     @annotate_tag
     def abstract_start(self, tag, attrs):
+        if (not self.metadata_registered and
+                tag == 'div' and 'slp' in self.classes(attrs)):
+            self.start_populating_textbuf()
+            return 'result_metadata'
         if tag == 'span' and 'st' in self.classes(attrs):
             self.start_populating_textbuf()
             return 'abstract_text'
 
     @retrieve_tag_annotation
     def abstract_end(self, tag, annotation):
-        if annotation == 'abstract_text':
+        if annotation == 'result_metadata':
+            self.stop_populating_textbuf()
+            self.metadata = self.pop_textbuf().strip().replace('\u200e', '')
+            self.metadata_registered = True
+        elif annotation == 'abstract_text':
             self.stop_populating_textbuf()
             self.abstract = self.pop_textbuf()
-            self.abstract_registered = False
+            self.abstract_registered = True
         elif annotation == 'abstract':
             self.set_handlers_to('result')
 
@@ -1385,7 +1411,7 @@ class GoogleParser(html.parser.HTMLParser):
         elif annotation == 'sitelink_title':
             self.set_handlers_to('sitelink')
 
-    ### Generic methods ###
+    # Generic methods
 
     # Set handle_starttag to SCOPE_start, and handle_endtag to SCOPE_end.
     def set_handlers_to(self, scope):
@@ -1601,7 +1627,7 @@ class Result(object):
         """Return a index-to-URL table for the current result.
 
         Normally, the table contains only a single entry, but when the result
-        contains sidelinks, all sidelinks are included in this table.
+        contains sitelinks, all sitelinks are included in this table.
 
         Returns
         -------
@@ -1633,6 +1659,7 @@ def require_keywords(method):
 
     return enforced_method
 
+
 def no_argument(method):
     # Normalize a do_* method of GooglerCmd that takes no argument to
     # one that takes an arg, but issue a warning when an nonempty
@@ -1647,6 +1674,7 @@ def no_argument(method):
 
     return enforced_method
 
+
 class GooglerCmd(object):
     """
     Command line interpreter and executor class for googler.
@@ -1819,7 +1847,7 @@ class GooglerCmd(object):
         """Show omniprompt and read user command line.
 
         Command line is always stripped, and each consecutive group of
-        whitespace is replaced with a single space charater. If the
+        whitespace is replaced with a single space character. If the
         command line is empty after stripping, when ignore it and keep
         reading. Exit with status 0 if we get EOF or an empty line
         (pre-strip, that is, a raw <enter>) twice in a row.
@@ -1873,15 +1901,42 @@ class GooglerCmd(object):
         # If > 5 results are being fetched each time,
         # block next when no parsed results in current fetch
         if not self.results and self._google_url._num > 5:
-            print('No results.', file=sys.stderr)
+            printerr('No results.')
         else:
             self._google_url.next_page()
             self.fetch_and_display()
 
     @require_keywords
-    @no_argument
-    def do_open(self):
-        open_url(self._google_url.full())
+    def do_open(self, *args):
+        if not args:
+            open_url(self._google_url.full())
+            return
+
+        for nav in args:
+            if nav == 'a':
+                for key, value in sorted(self._urltable.items()):
+                    open_url(self._urltable[key])
+            elif nav in self._urltable:
+                open_url(self._urltable[nav])
+            elif '-' in nav:
+                try:
+                    vals = [int(x) for x in nav.split('-')]
+                    if (len(vals) != 2):
+                        printerr('Invalid range %s.' % nav)
+                        continue
+
+                    if vals[0] > vals[1]:
+                        vals[0], vals[1] = vals[1], vals[0]
+
+                    for _id in range(vals[0], vals[1] + 1):
+                        if str(_id) in self._urltable:
+                            open_url(self._urltable[str(_id)])
+                        else:
+                            printerr('Invalid index %s.' % _id)
+                except ValueError:
+                    printerr('Invalid range %s.' % nav)
+            else:
+                printerr('Invalid index %s.' % nav)
 
     @require_keywords
     @no_argument
@@ -1921,7 +1976,7 @@ class GooglerCmd(object):
             # TODO: Automatic dispatcher
             #
             # We can't write a dispatcher for now because that could
-            # change behaviors of the prompt. However, we have already
+            # change behaviour of the prompt. However, we have already
             # laid a lot of ground work for the dispatcher, e.g., the
             # `no_argument' decorator.
             try:
@@ -1933,7 +1988,9 @@ class GooglerCmd(object):
                 elif cmd == 'n':
                     self.do_next('')
                 elif cmd == 'o':
-                    self.do_open('')
+                    self.do_open()
+                elif cmd.startswith('o '):
+                    self.do_open(*cmd[2:].split())
                 elif cmd == 'p':
                     self.do_previous('')
                 elif cmd == 'q':
@@ -1966,11 +2023,14 @@ class GooglerArgumentParser(argparse.ArgumentParser):
           n, p                  fetch the next or previous set of search results
           index                 open the result corresponding to index in browser
           f                     jump to the first page
-          o                     open the current search in browser
+          o [index|range|a ...] open space-separated result indices, numeric ranges
+                                (sitelinks unsupported in ranges), or all, in browser
+                                open the current search in browser, if no arguments
           g keywords            new Google search for 'keywords' with original options
+                                should be used to search omniprompt keys and indices
           q, ^D, double Enter   exit googler
           ?                     show omniprompt help
-          *                     any other string initiates a new search with original options
+          *                     other inputs issue a new search with original options
         """))
 
     # Print information on googler
@@ -1979,8 +2039,8 @@ class GooglerArgumentParser(argparse.ArgumentParser):
         file = sys.stderr if file is None else file
         file.write(textwrap.dedent("""
         Version %s
-        Copyright (C) 2008 Henri Hakkinen
-        Copyright (C) 2015-2016 Arun Prakash Jana <engineerarun@gmail.com>
+        Copyright © 2008 Henri Hakkinen
+        Copyright © 2015-2017 Arun Prakash Jana <engineerarun@gmail.com>
         Zhiming Wang <zmwangx@gmail.com>
         License: GPLv3
         Webpage: https://github.com/jarun/googler
@@ -2085,13 +2145,13 @@ def download_latest_googler(include_git=False):
         git_ref = response.read().decode('utf-8')
     else:
         # Get name of latest tag
-        request = urllib.request.Request('%s/tags?per_page=1' % API_REPO_BASE,
+        request = urllib.request.Request('%s/releases?per_page=1' % API_REPO_BASE,
                                          headers={'Accept': 'application/vnd.github.v3+json'})
         response = urllib.request.urlopen(request)
         if response.status != 200:
             raise http.client.HTTPException(response.reason)
         import json
-        git_ref = json.loads(response.read().decode('utf-8'))[0]['name']
+        git_ref = json.loads(response.read().decode('utf-8'))[0]['tag_name']
 
     # Download googler to a tempfile
     googler_download_url = '%s/%s/googler' % (RAW_DOWNLOAD_REPO_BASE, git_ref)
@@ -2209,11 +2269,12 @@ def parse_args(args=None, namespace=None):
 
     argparser = GooglerArgumentParser(description='Google from the command-line.')
     addarg = argparser.add_argument
-    addarg('-s', '--start', type=argparser.nonnegative_int, default=0, metavar='N',
-           help='start at the Nth result')
-    addarg('-n', '--count', dest='num', type=argparser.positive_int, default=10,
-           metavar='N', help='show N results (default 10)')
-    addarg('-N', '--news', action='store_true', help='show results from news section')
+    addarg('-s', '--start', type=argparser.nonnegative_int, default=0,
+           metavar='N', help='start at the Nth result')
+    addarg('-n', '--count', dest='num', type=argparser.positive_int,
+           default=10, metavar='N', help='show N results (default 10)')
+    addarg('-N', '--news', action='store_true',
+           help='show results from news section')
     addarg('-c', '--tld', metavar='TLD',
            help="""country-specific search with top-level domain .TLD, e.g., 'in'
            for India. Ref: https://en.wikipedia.org/wiki/List_of_Google_domains""")
@@ -2227,10 +2288,11 @@ def parse_args(args=None, namespace=None):
            help='set output colors (see man page for details)')
     addarg('-j', '--first', '--lucky', dest='lucky', action='store_true',
            help='open the first result in web browser and exit')
-    addarg('-t', '--time', dest='duration', type=argparser.is_duration, metavar='dN',
-           help='time limit search '
+    addarg('-t', '--time', dest='duration', type=argparser.is_duration,
+           metavar='dN', help='time limit search '
            '[h5 (5 hrs), d5 (5 days), w5 (5 weeks), m5 (5 months), y5 (5 years)]')
-    addarg('-w', '--site', metavar='SITE', help='search a site using Google')
+    addarg('-w', '--site', dest='sites', action='append', metavar='SITE',
+           help='search a site using Google')
     addarg('-p', '--proxy', default=https_proxy_from_environment(),
            help='tunnel traffic through an HTTPS proxy (HOST:PORT)')
     addarg('--noua', action='store_true', help='disable user agent')
@@ -2238,7 +2300,7 @@ def parse_args(args=None, namespace=None):
            help='disable TCP optimizations and forced TLS 1.2')
     addarg('--json', action='store_true',
            help='output in JSON format; implies --exact and --noprompt')
-    addarg('--enable-browser-output', action='store_true',
+    addarg('--show-browser-logs', action='store_true',
            help='do not suppress browser output (stdout and stderr)')
     addarg('--np', '--noprompt', dest='noninteractive', action='store_true',
            help='search and exit, do not prompt')
@@ -2292,10 +2354,11 @@ def main():
         GooglerCmd.colors = colors
 
         # Handle browser output suppression
-        if opts.enable_browser_output:
+        if opts.show_browser_logs:
             open_url.suppress_browser_output = False
 
         if opts.noua:
+            logger.debug('User Agent is disabled')
             ua = False
 
         repl = GooglerCmd(opts)
@@ -2315,7 +2378,7 @@ def main():
             # Interactive mode
             repl.cmdloop()
     except Exception as e:
-        # If debugging mode is enabled, let the exception through for a traceback;
+        # With debugging on, let the exception through for a traceback;
         # otherwise, only print the exception error message.
         if logger.isEnabledFor(logging.DEBUG):
             raise
